@@ -9,26 +9,28 @@ import itertools
 import json
 import collections
 
+from copy import deepcopy
 from statistics import mean
 from abc import ABC, abstractmethod
 from itertools import product, islice, accumulate, groupby
 from statistics import median
 from pathlib import Path
-from typing import (
-    Iterable, Tuple, Union, Sequence, Callable, 
-    Generic, TypeVar, Dict, Any, cast, Optional, overload
-)
+from typing import Iterable, Tuple, Union, Sequence, Generic, TypeVar, Dict, Any, cast, Optional, overload, List
 
-from coba.simulations import LazySimulation, JsonSimulation, Simulation, Context, Action
-from coba.learners import Learner
+from coba.learners import Learner, Choice, Key
+from coba.simulations import JsonSimulation, Simulation, Context, Action, Reward
 from coba.execution import ExecutionContext
-from coba.data import Pipe, MemorySink, MemorySource, StopPipe, Filter, DiskSource, DiskSink, JsonEncode, JsonDecode, Table
-from coba.random import Random
+from coba.data.structures import Table
+from coba.data.filters import Filter, JsonEncode, JsonDecode
+from coba.data.sources import Source, MemorySource, DiskSource
+from coba.data.sinks import Sink, MemorySink, DiskSink
+from coba.data.pipes import Pipe, StopPipe
+from coba.statistics import OnlineMean, OnlineVariance
+from coba.utilities import check_matplotlib_support, check_pandas_support
+from coba.random import CobaRandom
 
-_C_out = TypeVar('_C_out', bound=Context, covariant=True)
-_A_out = TypeVar('_A_out', bound=Action, covariant=True)
-_C     = TypeVar('_C'    , bound=Context)
-_A     = TypeVar('_A'    , bound=Action)
+_C = TypeVar('_C', bound=Context)
+_A = TypeVar('_A', bound=Action)
 
 class Result:
     """A class for creating and returning the result of a Benchmark evaluation."""
@@ -86,12 +88,136 @@ class Result:
         )
 
     def to_pandas(self) -> Tuple[Any,Any,Any]:
+
+        check_pandas_support("Result.to_pandas")
+
         l = self.learners.to_pandas()
         s = self.simulations.to_pandas()
         b = self.batches.to_pandas()
 
         return (l,s,b)
 
+    def standard_plot(self, select_learners: Sequence[int] = None,  show_err: bool = False, show_sd: bool = False) -> None:
+
+        check_matplotlib_support('Plots.standard_plot')
+
+        def _plot(axes, label, xs, ys, vs, ns):
+            axes.plot(xs, ys, label=label)
+
+            if show_sd:
+                ls = [ y-math.sqrt(v) for y,v in zip(ys,vs) ]
+                us = [ y+math.sqrt(v) for y,v in zip(ys,vs) ]
+                axes.fill_between(xs, ls, us, alpha = 0.1)
+
+            if show_err:
+                # I don't really understand what this is... For each x our distribution
+                # is changing so its VAR is also changing. What does it mean to calculate
+                # sample variance from a deterministic collection of random variables with
+                # different distributions? For example sample variance of 10 random variables
+                # from dist1 and 10 random variables from dist2... This is not the same as 20
+                # random variables with 50% chance drawing from dist1 and 50% chance of drawing
+                # from dist2. So the distribution can only be defined over the whole space (i.e.,
+                # all 20 random variables) and not for a specific random variable. Oh well, for
+                # now I'm leaving this as it is since I don't have any better ideas. I think what
+                # I've done is ok, but I need to more some more thought into it.
+                ls = [ y-math.sqrt(v/n) for y,v,n in zip(ys,vs,ns) ]
+                us = [ y+math.sqrt(v/n) for y,v,n in zip(ys,vs,ns) ]
+                axes.fill_between(xs, ls, us, alpha = 0.1)
+
+        learners, _, batches = self.to_indexed_tuples()
+
+        learners = {key:value for key,value in learners.items() if select_learners is None or key in select_learners}
+        batches  = {key:value for key,value in batches.items() if select_learners is None or value.learner_id in select_learners}
+
+        learner_index_key = lambda batch: (batch.learner_id, batch.batch_index)
+        sorted_batches    = sorted(batches.values(), key=learner_index_key)
+        grouped_batches   = groupby(groupby(sorted_batches , key=learner_index_key), key=lambda x: x[0][0])
+
+        max_batch_N = 0
+
+        indexes     = cast(Dict[int,List[int  ]], collections.defaultdict(list))
+        incounts    = cast(Dict[int,List[int  ]], collections.defaultdict(list))
+        inmeans     = cast(Dict[int,List[float]], collections.defaultdict(list))
+        invariances = cast(Dict[int,List[float]], collections.defaultdict(list))
+        cucounts    = cast(Dict[int,List[int  ]], collections.defaultdict(list))
+        cumeans     = cast(Dict[int,List[float]], collections.defaultdict(list))
+        cuvariances = cast(Dict[int,List[float]], collections.defaultdict(list))
+
+        for learner_id, learner_batches in grouped_batches:
+
+            cucount    = 0
+            cumean     = OnlineMean()
+            cuvariance = OnlineVariance()
+
+            for (_, batch_index), index_batches in learner_batches:
+
+                incount    = 0
+                inmean     = OnlineMean()
+                invariance = OnlineVariance()
+
+                for N, reward in [ (b.N, b.reward) for b in index_batches]:
+                    
+                    max_batch_N = max(N, max_batch_N)
+                    
+                    incount     = incount + 1
+                    inmean      .update(reward)
+                    invariance  .update(reward)
+                    cucount     = cucount + 1
+                    cumean      .update(reward)
+                    cuvariance  .update(reward)
+
+                #sanity check, sorting above (in theory) should take care of this...
+                #if this isn't the case then the cu* values will be incorrect...
+                assert indexes[learner_id] == [] or batch_index > indexes[learner_id][-1]
+
+                incounts[learner_id].append(incount)
+                indexes[learner_id].append(batch_index)
+                inmeans[learner_id].append(inmean.mean)
+                invariances[learner_id].append(invariance.variance)
+                cucounts[learner_id].append(cucount)
+                cumeans[learner_id].append(cumean.mean)
+                cuvariances[learner_id].append(cuvariance.variance)
+
+        import matplotlib.pyplot as plt #type: ignore
+
+        fig = plt.figure(figsize=(12,4))
+
+        index_unit = "Interaction" if max_batch_N ==1 else "Batch"
+        
+        ax1 = fig.add_subplot(1,2,1) #type: ignore
+        ax2 = fig.add_subplot(1,2,2) #type: ignore
+
+        for learner_id in learners:
+            _plot(ax1, learners[learner_id].full_name, indexes[learner_id], inmeans[learner_id], invariances[learner_id], incounts[learner_id])
+
+        ax1.set_title(f"Instantaneous Reward")
+        ax1.set_ylabel("Reward")
+        ax1.set_xlabel(f"{index_unit} Index")
+
+        for learner_id in learners:
+            _plot(ax2, learners[learner_id].full_name, indexes[learner_id], cumeans[learner_id], cuvariances[learner_id], cucounts[learner_id])
+
+        ax2.set_title("Progressive Validation")
+        #ax2.set_ylabel("Reward")
+        ax2.set_xlabel(f"{index_unit} Index")
+
+        (bot1, top1) = ax1.get_ylim()
+        (bot2, top2) = ax2.get_ylim()
+
+        ax1.set_ylim(min(bot1,bot2), max(top1,top2))
+        ax2.set_ylim(min(bot1,bot2), max(top1,top2))
+
+        scale = 0.75
+        box1 = ax1.get_position()
+        box2 = ax2.get_position()
+        ax1.set_position([box1.x0, box1.y0 + box1.height * (1-scale), box1.width, box1.height * scale])
+        ax2.set_position([box2.x0, box2.y0 + box2.height * (1-scale), box2.width, box2.height * scale])
+
+        # Put a legend below current axis
+        fig.legend(*ax1.get_legend_handles_labels(), loc='upper center', bbox_to_anchor=(.5, .175), ncol=2) #type: ignore
+
+        plt.show()
+
     def __str__(self) -> str:
         return str({ "Learners": len(self.learners), "Simulations": len(self.simulations), "Batches": len(self.batches) })
 
@@ -208,25 +334,9 @@ class Transaction:
         return ["L", learner_id, kwargs]
 
     @staticmethod
-    def learners(factories: 'Sequence[LearnerFactory]') -> Iterable[Any]:
-        for index, factory in enumerate(factories):
-                learner = factory.create()
-                try:
-                    family = learner.family
-                except:
-                    family = learner.__class__.__name__
-
-                try:
-                    params = learner.params
-                except:
-                    params =  {}
-
-                if len(params) > 0:
-                    full_name = f"{family}({','.join(f'{k}={v}' for k,v in params.items())})"
-                else:
-                    full_name = family
-
-                yield Transaction.learner(index, family=family, full_name=full_name, **params)
+    def learners(learners: 'Sequence[BenchmarkLearner[_C, _A]]') -> Iterable[Any]:
+        for index, learner in enumerate(learners):
+            yield Transaction.learner(index, family=learner.family, full_name=learner.full_name, **learner.params)
 
     @staticmethod
     def simulation(simulation_id: int, **kwargs) -> Any:
@@ -250,6 +360,39 @@ class Transaction:
         """
         return ["B", (learner_id, simulation_id, seed, batch_index), kwargs]
 
+class TaskSource(Source):
+    
+    def __init__(self, simulations, seeds, learners, restored) -> None:
+        self._simulations = simulations
+        self._seeds       = seeds
+        self._learners    = learners
+        self._restored    = restored
+
+    def read(self) -> Iterable:
+        simulations = dict(enumerate(self._simulations))
+        seeds       = self._seeds
+        learners    = dict(enumerate(self._learners))
+        restored    = self._restored
+
+        expected_batch_count = cast(Dict[Tuple[Any,Any,Any], int], collections.defaultdict(lambda:  int(-1)))
+        restored_batch_count = cast(Dict[Tuple[Any,Any,Any], int], collections.defaultdict(lambda:  int( 0)))
+
+        for k,v in restored.simulations.rows.items():
+            expected_batch_count[k] = v[restored.simulations._columns.index('batch_count')]
+
+        for k in restored.batches.rows:
+            restored_batch_count[(k[1],k[2],k[0])] +=1
+
+        is_not_complete = lambda t: restored_batch_count[t] != expected_batch_count[t[0]]
+        
+        task_keys       = filter(is_not_complete, product(simulations, seeds, learners))
+        task_key_groups = groupby(task_keys, key=lambda t: t[0])
+
+        seed_learner = lambda t: (t[1], t[2], learners[t[2]])
+        
+        for simulation_key, task_key_group in task_key_groups:
+            yield simulation_key, simulations[simulation_key], list(map(seed_learner, task_key_group))
+    
 class TaskToTransactions(Filter):
 
     def __init__(self, ignore_first: bool, ignore_raise: bool, batcher: Batcher) -> None:
@@ -268,7 +411,7 @@ class TaskToTransactions(Filter):
         simulation       = task[1]
 
         try:
-            with self._lazy_simulation(simulation) as simulation:
+            with WithSimulation(simulation) as simulation:
 
                 batch_sizes = self._batcher.batch_sizes(len(simulation.interactions))
 
@@ -278,8 +421,11 @@ class TaskToTransactions(Filter):
                     context_size      = int(median(self._context_sizes(simulation))),
                     action_count      = int(median(self._action_counts(simulation))))
 
-                for (seed, learner_index, factory) in task[2]:
-                    learner = factory.create()
+                for (seed, learner_index, learner_template) in task[2]:
+
+                    learner = deepcopy(learner_template)
+                    learner.init()
+
                     for batch_index, batch in enumerate(self._shuffle_batch(simulation.interactions, seed, batch_sizes)):
                         for transaction in self._process_batch(simulation_index, simulation, seed, learner_index, learner, batch_index, batch):
                             yield transaction
@@ -319,16 +465,13 @@ class TaskToTransactions(Filter):
         if batch_index >= 0:
             yield Transaction.batch(learner_index, simulation_index, seed, batch_index, N=len(rewards), reward=round(mean(rewards),5))
 
-    def _shuffle_batch(self, interactions, seed, batch_sizes):
-        batch_slices = list(accumulate([0] + list(batch_sizes)))
-        interactions = Random(seed).shuffle(interactions) if seed else interactions
+    def _shuffle_batch(self, interactions, seed, batch_sizes: Sequence[int]):
+        batch_slices = list(accumulate([0] + list(batch_sizes))) #type: ignore
+        interactions = CobaRandom(seed).shuffle(interactions) if seed else interactions
         
         for i in range(len(batch_slices)-1):
             yield islice(interactions, batch_slices[i], batch_slices[i+1])
 
-    def _lazy_simulation(self, simulation: Simulation) -> Union[LazySimulation,JsonSimulation]:
-        return simulation if isinstance(simulation, (LazySimulation, JsonSimulation)) else LazySimulation(lambda: simulation)
-
     def _context_sizes(self, simulation: Simulation) -> Iterable[int]:
         for context in [i.context for i in simulation.interactions]:
             yield 0 if context is None else len(context) if isinstance(context,tuple) else 1
@@ -427,53 +570,101 @@ class TransactionIsNew(Filter):
 
             yield item
 
-class LearnerFactory(Generic[_C_out, _A_out]):
-    def __init__(self, ctor: Callable[...,Learner[_C_out,_A_out]], *args, **kwargs) -> None:
-        self._ctor   = ctor
-        self._args   = args
-        self._kwargs = kwargs
+class TransactionSink(Sink):
+    
+    def __init__(self, transaction_log: Optional[str], restored: Result) -> None:
+        self._sink = Pipe.join([JsonEncode()], DiskSink(transaction_log)) if transaction_log else MemorySink()
+        self._sink = Pipe.join([TransactionIsNew(restored)], self._sink)
 
-    def create(self) -> Learner[_C_out,_A_out]:
-        return self._ctor(*self._args, **self._kwargs)
+    def write(self, items: Sequence[Any]) -> None:
+        self._sink.write(items)
 
-class Benchmark(Generic[_C,_A], ABC):
-    """The interface for Benchmark implementations."""
+    @property
+    def result(self) -> Result:
+        if isinstance(self._sink, Pipe.FiltersSink):
+            final_sink = self._sink.final_sink()
+        else:
+            final_sink = self._sink
 
-    @abstractmethod
-    def evaluate(self, factories: Sequence[LearnerFactory[_C,_A]]) -> Result:
-        """Calculate the performance for a provided bandit Learner.
+        if isinstance(final_sink, MemorySink):
+            return Result.from_transactions(cast(Iterable[Any], final_sink.items))
 
-        Args:
-            factories: A sequence of functions to create Learner instances. Each function 
-                should always create the same Learner in order to get an unbiased performance 
-                Result. This method can be as simple as `lambda: MyLearner(...)`.
+        if isinstance(final_sink, DiskSink):
+            return Result.from_transaction_log(final_sink.filename)
 
-        Returns:
-            The resulting performance statistics for each given learner to evaluate.
+        raise Exception("Transactions were written to an unrecognized sink.")
 
-        Remarks:
-            The learner factory is necessary because a Result can be calculated using
-            observed performance over several simulations. In these cases the easiest 
-            way to reset a learner's learned policy is to create a new learner.
-        """
-        ...
+class WithSimulation:
+    def __init__(self, simulation: Simulation) -> None:
+        self._simulation = simulation
 
-class UniversalBenchmark(Benchmark[_C,_A]):
+    def __enter__(self) -> Simulation:
+        with ExecutionContext.Logger.log(f"loading simulation..."):
+            try:
+                return self._simulation.__enter__() #type: ignore
+            except AttributeError:
+                return self._simulation
+
+    def __exit__(self, exception_type, exception_value, traceback) -> None:
+        try:
+            self._simulation.__exit__(exception_type, exception_value, traceback) #type: ignore
+        except AttributeError:
+            pass
+
+class BenchmarkLearner(Learner[_C,_A]):
+    
+    @property
+    def family(self) -> str:
+        try:
+            return self._learner.family
+        except AttributeError:
+            return self._learner.__class__.__name__
+
+    @property
+    def params(self) -> Dict[str, Any]:
+        try:
+            return self._learner.params
+        except AttributeError:
+            return {}
+    
+    @property
+    def full_name(self) -> str:
+        if len(self.params) > 0:
+            return f"{self.family}({','.join(f'{k}={v}' for k,v in self.params.items())})"
+        else:
+            return self.family
+
+    def __init__(self, learner: Learner[_C,_A]) -> None:
+        self._learner = learner
+
+    def init(self) -> None:
+        try:
+            self._learner.init()
+        except AttributeError:
+            pass
+
+    def choose(self, key: Key, context: _C, actions: Sequence[_A]) -> Choice:
+        return self._learner.choose(key, context, actions)
+    
+    def learn(self, key: Key, context: _C, action: _A, reward: Reward) -> None:
+        self._learner.learn(key, context, action, reward)
+
+class Benchmark(Generic[_C,_A]):
     """An on-policy Benchmark using samples drawn from simulations to estimate performance statistics."""
 
     @staticmethod
-    def from_file(filename:str) -> 'UniversalBenchmark[Context,Action]':
+    def from_file(filename:str) -> 'Benchmark[Context,Action]':
         """Instantiate a Benchmark from a config file."""
 
         suffix = Path(filename).suffix
         
         if suffix == ".json":
-            return UniversalBenchmark.from_json(Path(filename).read_text())
+            return Benchmark.from_json(Path(filename).read_text())
 
         raise Exception(f"The provided file type ('{suffix}') is not a valid format for benchmark configuration")
 
     @staticmethod
-    def from_json(json_val:Union[str, Dict[str,Any]]) -> 'UniversalBenchmark[Context,Action]':
+    def from_json(json_val:Union[str, Dict[str,Any]]) -> 'Benchmark[Context,Action]':
         """Create a UniversalBenchmark from json text or object.
 
         Args:
@@ -499,7 +690,7 @@ class UniversalBenchmark(Benchmark[_C,_A]):
         ignore_raise = config.get("ignore_raise", True)
         shuffle      = config.get("shuffle", [None])
 
-        return UniversalBenchmark(simulations, batcher, ignore_first=ignore_first, ignore_raise=ignore_raise, shuffle_seeds=shuffle)
+        return Benchmark(simulations, batcher, ignore_first=ignore_first, ignore_raise=ignore_raise, shuffle_seeds=shuffle)
 
     @overload
     def __init__(self, 
@@ -589,23 +780,23 @@ class UniversalBenchmark(Benchmark[_C,_A]):
         self._processes        = kwargs.get('processes', None)
         self._maxtasksperchild = kwargs.get('maxtasksperchild', None)
 
-    def ignore_raise(self, value:bool=True) -> 'UniversalBenchmark[_C,_A]':
+    def ignore_raise(self, value:bool=True) -> 'Benchmark[_C,_A]':
         self._ignore_raise = value
         return self
 
-    def ignore_first(self, value:bool=True) -> 'UniversalBenchmark[_C,_A]':
+    def ignore_first(self, value:bool=True) -> 'Benchmark[_C,_A]':
         self._ignore_first = value
         return self
 
-    def processes(self, value:int) -> 'UniversalBenchmark[_C,_A]':
+    def processes(self, value:int) -> 'Benchmark[_C,_A]':
         self._processes = value
         return self
 
-    def maxtasksperchild(self, value:int) -> 'UniversalBenchmark[_C,_A]':
+    def maxtasksperchild(self, value:int) -> 'Benchmark[_C,_A]':
         self._maxtasksperchild = value
         return self
 
-    def evaluate(self, factories: Sequence[LearnerFactory[_C,_A]], transaction_log:str = None) -> Result:
+    def evaluate(self, learners: Sequence[Learner[_C,_A]], transaction_log:str = None) -> Result:
         """Collect observations of a Learner playing the benchmark's simulations to calculate Results.
 
         Args:
@@ -614,13 +805,13 @@ class UniversalBenchmark(Benchmark[_C,_A]):
         Returns:
             See the base class for more information.
         """
-
+        bench_learners       = [ BenchmarkLearner(learner) for learner in learners ]
         restored             = Result.from_transaction_log(transaction_log)
-        task_source          = MemorySource(self._make_tasks(self._simulations, self._seeds, factories, restored))
+        task_source          = TaskSource(self._simulations, self._seeds, bench_learners, restored)
         task_to_transactions = TaskToTransactions(self._ignore_first, self._ignore_raise, self._batcher)
-        transaction_sink     = Pipe.join([JsonEncode()], DiskSink(transaction_log)) if transaction_log else MemorySink()
+        transaction_sink     = TransactionSink(transaction_log, restored)
 
-        n_given_learners    = len(factories)
+        n_given_learners    = len(bench_learners)
         n_given_simulations = len(self._simulations)
         n_given_seeds       = len(self._seeds)
         given_batcher       = self._batcher.__class__.__name__
@@ -636,47 +827,12 @@ class UniversalBenchmark(Benchmark[_C,_A]):
         preamble_transactions = []
         preamble_transactions.append(Transaction.version(TransactionPromote.CurrentVersion))
         preamble_transactions.append(Transaction.benchmark(n_given_learners, n_given_simulations, n_given_seeds, given_batcher, ignore_first))
-        preamble_transactions.extend(Transaction.learners(factories))
+        preamble_transactions.extend(Transaction.learners(bench_learners))
 
         mp = self._processes if self._processes else ExecutionContext.Config.processes
         mt = self._maxtasksperchild if self._maxtasksperchild else ExecutionContext.Config.maxtasksperchild
-
-        transaction_sink = Pipe.join([TransactionIsNew(restored)], transaction_sink)
-        Pipe.join(MemorySource(preamble_transactions), [], transaction_sink).run()
-        Pipe.join(task_source, [task_to_transactions], transaction_sink).run(mp,mt)
-
-        if isinstance(transaction_sink, Pipe.FiltersSink):
-            transaction_sink = transaction_sink.final_sink()
-
-        if isinstance(transaction_sink, MemorySink):
-            return Result.from_transactions(transaction_sink.items)
-
-        if isinstance(transaction_sink, DiskSink):
-            return Result.from_transaction_log(transaction_sink.filename)
-
-        raise Exception("Transactions were written to an unrecognized sink.")
-
-    def _make_tasks(self, simulations, seeds, factories, restored) -> Iterable[Any]:
-
-        simulations = dict(enumerate(simulations))
-        seeds       = seeds
-        factories   = dict(enumerate(factories))
-
-        expected_batch_count = cast(Dict[Tuple[Any,Any,Any], int], collections.defaultdict(lambda:  int(-1)))
-        restored_batch_count = cast(Dict[Tuple[Any,Any,Any], int], collections.defaultdict(lambda:  int( 0)))
-
-        for k,v in restored.simulations.rows.items():
-            expected_batch_count[k] = v[restored.simulations._columns.index('batch_count')]
-
-        for k in restored.batches.rows:
-            restored_batch_count[(k[1],k[2],k[0])] +=1
-
-        is_not_complete = lambda t: restored_batch_count[t] != expected_batch_count[t[0]]
         
-        task_keys       = filter(is_not_complete, product(simulations, seeds, factories))
-        task_key_groups = groupby(task_keys, key=lambda t: t[0])
+        Pipe.join(MemorySource(preamble_transactions), []                    , transaction_sink).run(1,None)
+        Pipe.join(task_source                        , [task_to_transactions], transaction_sink).run(mp,mt)
 
-        seed_learner = lambda t: (t[1], t[2], factories[t[2]])
-        
-        for simulation_key, task_key_group in task_key_groups:
-            yield simulation_key, simulations[simulation_key], list(map(seed_learner, task_key_group))
\ No newline at end of file
+        return transaction_sink.result
\ No newline at end of file
