@@ -12,16 +12,16 @@ import sys
 import os
 import traceback
 
+from io import UnsupportedOperation
 from contextlib import contextmanager
 from itertools import repeat
 from gzip import compress, decompress
 from abc import ABC, abstractmethod
-from io import BytesIO
 from pathlib import Path
 from datetime import datetime
 from typing import (
-    Callable, ContextManager, Union, Generic, TypeVar, Dict, 
-    IO, Mapping, Any, Optional, List, MutableMapping, cast, Iterator,
+    Callable, ContextManager, Union, Generic, TypeVar, Dict, IO,
+    Mapping, Any, Optional, List, MutableMapping, cast, Iterator
 )
 
 _K = TypeVar("_K")
@@ -31,8 +31,8 @@ class TemplatingEngine:
     """This class materializes templates within benchmark json files.
     
     The templating engine works as follows: 
-        1. Look in the root object for a "templates" object. Templates should be be objects themselves
-        with hard coded values and variables. Variable values are indicated by beginning with a $.
+        1. Look in the root object for a "templates" key. Templates should be objects themselves
+        with constants and variables. Variable values are indicated by beginning with a $.
         2. Recursively walk through the remainder of the children from the root. For every object found
         check to see if it has a "template" value. 
         3. For every object found with a "template" value defined materialize all of that template's static
@@ -112,23 +112,10 @@ class CobaConfig():
                     config = json.load(fs)
                 break
 
-        self._config = config
-
-    @property
-    def openml_api_key(self) -> Optional[str]:
-        return self._config.get("openml_api_key", None)
-
-    @property
-    def file_cache(self) -> Mapping[str,Any]:
-        return self._config.get("file_cache", {"type":"none"})
-
-    @property
-    def processes(self) -> int:
-        return self._config.get("processes", 1)
-    
-    @property
-    def maxtasksperchild(self) -> Optional[int]:
-        return self._config.get("maxtasksperchild", None)
+        self.openml_api_key   = config.get("openml_api_key", None)
+        self.file_cache       = config.get("file_cache", {"type":"none"})
+        self.processes        = config.get("processes", 1)
+        self.maxtasksperchild = config.get("maxtasksperchild", None)
 
 class CacheInterface(Generic[_K, _V], ABC):
     """The interface for a cacher."""
@@ -384,41 +371,45 @@ def redirect_stderr(to: IO[str]):
         Python modules such as `contextlib.redirect_stdout` and `contextlib.redirect_stderr`. See
         https://docs.python.org/3/library/contextlib.html#contextlib.redirect_stdout for more info.
     """
-    
-    #we assume that this fd is the same
-    #one that is used by our C library
-    stderr_fd = sys.stderr.fileno()
-
-    def _redirect_stderr(redirect_stderr_fd):
-        
-        #first we flush Python's stderr. It should be noted that this
-        #doesn't close the file descriptor (i.e., sys.stderr.fileno())
-        #or Python's wrapper around the stderr_fd.
-        sys.stderr.flush()
-    
-        # next we change the stderr_fd to point to the
-        # file contained in the redirect_stderr_fd.
-        # If C has anything buffered for stderr it
-        # will now go to the new fd. There do appear
-        # to be ways to flush C buffers from Python 
-        # but I'm not sure it is worth it given the
-        # amount of complexity it adds to the code.
-        # This change also means that sys.stderr now
-        # points to a new file since sys.stderr points
-        # to whatever file is at stderr_fd
-        os.dup2(redirect_stderr_fd, stderr_fd)
-
-    # when we dup there are now two fd's
-    # pointing to the same file. Closing
-    # one of these doesn't close the other.
-    # therefore it is on us to close the
-    # duplicate fd we make here before ending.
-    old_stderr_fd = os.dup(stderr_fd)
-    new_stderr_fd = to.fileno()
-
     try:
-        _redirect_stderr(new_stderr_fd)
-        yield # allow code to be run with the redirected stderr
-    finally:
-        _redirect_stderr(old_stderr_fd) 
-        os.close(old_stderr_fd)
+        #we assume that this fd is the same
+        #one that is used by our C library
+        stderr_fd = sys.stderr.fileno()
+
+        def _redirect_stderr(redirect_stderr_fd):
+            
+            #first we flush Python's stderr. It should be noted that this
+            #doesn't close the file descriptor (i.e., sys.stderr.fileno())
+            #or Python's wrapper around the stderr_fd.
+            sys.stderr.flush()
+        
+            # next we change the stderr_fd to point to the
+            # file contained in the redirect_stderr_fd.
+            # If C has anything buffered for stderr it
+            # will now go to the new fd. There do appear
+            # to be ways to flush C buffers from Python 
+            # but I'm not sure it is worth it given the
+            # amount of complexity it adds to the code.
+            # This change also means that sys.stderr now
+            # points to a new file since sys.stderr points
+            # to whatever file is at stderr_fd
+            os.dup2(redirect_stderr_fd, stderr_fd)
+
+        # when we dup there are now two fd's
+        # pointing to the same file. Closing
+        # one of these doesn't close the other.
+        # therefore it is on us to close the
+        # duplicate fd we make here before ending.
+        old_stderr_fd = os.dup(stderr_fd)
+        new_stderr_fd = to.fileno()
+
+        try:
+            _redirect_stderr(new_stderr_fd)
+            yield # allow code to be run with the redirected stderr
+        finally:
+            _redirect_stderr(old_stderr_fd) 
+            os.close(old_stderr_fd)
+    except UnsupportedOperation:
+        #if for some reason we weren't able to redirect
+        #then simply move on. No reason to stop working.
+        yield
\ No newline at end of file
