@@ -8,19 +8,21 @@ import math
 import collections
 
 from abc import ABC, abstractmethod
-from typing import Any, Callable, Sequence, Tuple, Optional, Dict, cast, Generic, TypeVar, overload, Union, List, Type
+from typing import Any, Sequence, Tuple, Optional, Dict, cast, Generic, TypeVar, overload, Union, Callable
 from collections import defaultdict
-from inspect import signature
 
-import coba.random
 import coba.vowpal as VW
 
+from coba.random import CobaRandom
 from coba.simulations import Context, Action, Reward, Choice, Key
 from coba.statistics import OnlineVariance
 
 _C_in = TypeVar('_C_in', bound=Context, contravariant=True)
 _A_in = TypeVar('_A_in', bound=Action , contravariant=True)
 
+_C_out = TypeVar('_C_out', bound=Context, covariant=True)
+_A_out = TypeVar('_A_out', bound=Action , covariant=True)
+
 class Learner(Generic[_C_in, _A_in], ABC):
     """The interface for Learner implementations."""
 
@@ -42,6 +44,10 @@ class Learner(Generic[_C_in, _A_in], ABC):
         """
         ...
 
+    def init(self) -> None:
+        """An optional initialization method called once after pickling."""        
+        pass
+
     @abstractmethod
     def choose(self, key: Key, context: _C_in, actions: Sequence[_A_in]) -> Choice:
         """Choose which action to take.
@@ -91,98 +97,20 @@ class Learner(Generic[_C_in, _A_in], ABC):
         """
         ...
 
-class LambdaLearner(Learner[_C_in, _A_in]):
-    """A Learner implementation that chooses and learns according to provided lambda functions."""
+class LearnerFactory(Generic[_C_out, _A_out]):
+    def __init__(self, ctor: Callable[...,Learner[_C_out,_A_out]], *args, **kwargs) -> None:
+        self._ctor   = ctor
+        self._args   = args
+        self._kwargs = kwargs
 
-    @overload
-    def __init__(self, 
-                choose: Callable[[_C_in, Sequence[_A_in]], Choice], 
-                learn : Optional[Callable[[_C_in, _A_in, Reward],None]] = None,
-                family: str = "Lambda",
-                params: Dict[str,Any] = {}) -> None:
-        ...
-    
-    @overload
-    def __init__(self, 
-                 choose: Callable[[Key, _C_in, Sequence[_A_in]], Choice], 
-                 learn : Optional[Callable[[Key, _C_in, _A_in, Reward],None]] = None,
-                 family: str = "Lambda",
-                 params: Dict[str,Any] = {}) -> None:
-        ...
-
-    def __init__(self, choose, learn = None, family: str = "Lambda", params: Dict[str,Any] = {}) -> None:
-        """Instantiate LambdaLearner.
-
-        Args:
-            chooser: A function matching the `super().choose()` signature. All parameters are passed straight through.
-            learner: A function matching the `super().learn()` signature. If provided all parameters are passed
-                straight through. If the function isn't provided then no learning occurs.
-            family: The family that the lambda learner belongs to.
-            params: The parameters used when creating the lambda learner.
-        """
-
-        if len(signature(choose).parameters) == 2:
-            og_choose = choose
-            choose = lambda k,c,A: cast(Callable[[_C_in, Sequence[_A_in]], Choice], og_choose)(c,A)
-
-        if learn is not None and len(signature(learn).parameters) == 3:
-            og_learn = learn
-            learn = lambda k,c,a,r: cast(Callable[[_C_in, _A_in, Reward], None], og_learn)(c,a,r)
-
-        self._choose = choose
-        self._learn  = learn
-        self._family = family
-        self._params = params
-
-    @property
-    def family(self) -> str:
-        """The family of the learner.
-        
-        See the base class for more information
-        """
-        return self._family
-
-    @property
-    def params(self) -> Dict[str, Any]:
-        """The parameters of the learner.
-        
-        See the base class for more information
-        """
-        return self._params
-
-    def choose(self, key: Key, context: _C_in, actions: Sequence[_A_in]) -> Choice:
-        """Choose via the provided lambda function.
-
-        Args:
-            key: The key identifying the interaction we are choosing for.
-            context: The context we're currently in. See the base class for more information.
-            actions: The actions to choose from. See the base class for more information.
-
-        Returns:
-            The index of the selected action. See the base class for more information.
-        """
-
-        return self._choose(key, context, actions)
-
-    def learn(self, index: int, context: _C_in, action: _A_in, reward: Reward) -> None:
-        """Learn via the optional lambda function or learn nothing without a lambda function.
-
-        Args:
-            key: The key identifying the interaction this observed reward came from.
-            context: The context we're learning about. See the base class for more information.
-            action: The action that was selected in the context. See the base class for more information.
-            reward: The reward that was gained from the action. See the base class for more information.
-        """
-        if self._learn is None:
-            pass
-        else:
-            self._learn(index, context,action,reward)
+    def create(self) -> Learner[_C_out,_A_out]:
+        return self._ctor(*self._args, **self._kwargs)
 
 class RandomLearner(Learner[Context, Action]):
     """A Learner implementation that selects an action at random and learns nothing."""
 
     def __init__(self, seed: Optional[int] = None):
-        self._random = coba.random.Random(seed)
+        self._random = CobaRandom(seed)
 
     @property
     def family(self) -> str:
@@ -242,7 +170,7 @@ class EpsilonLearner(Learner[Context, Action]):
 
         self._epsilon         = epsilon
         self._include_context = include_context
-        self._random          = coba.random.Random(seed)
+        self._random          = CobaRandom(seed)
 
         self._N: Dict[Tuple[Context, Action], int            ] = defaultdict(int)
         self._Q: Dict[Tuple[Context, Action], Optional[float]] = defaultdict(int)
@@ -310,6 +238,8 @@ class EpsilonLearner(Learner[Context, Action]):
 class UcbTunedLearner(Learner[Context, Action]):
     """This is an implementation of Auer et al. (2002) UCB1-Tuned algorithm.
 
+    This algorithm assumes that the reward distribution has support in [0,1].
+
     References:
         Auer, Peter, Nicolo Cesa-Bianchi, and Paul Fischer. "Finite-time analysis of 
         the multiarmed bandit problem." Machine learning 47.2-3 (2002): 235-256.
@@ -319,11 +249,11 @@ class UcbTunedLearner(Learner[Context, Action]):
 
         self._init_a: int = 0
         self._t     : int = 0
-        self._s     : Dict[Action,int] = {}
-        self._m     : Dict[Action,float] = {}
-        self._v     : Dict[Action,OnlineVariance] = defaultdict(OnlineVariance)
+        self._s     : Dict[Action, int           ] = defaultdict(int)
+        self._m     : Dict[Action, float         ] = {}
+        self._v     : Dict[Action, OnlineVariance] = defaultdict(OnlineVariance)
         
-        self._random = coba.random.Random(seed)
+        self._random = CobaRandom(seed)
 
     @property
     def family(self) -> str:
@@ -372,10 +302,6 @@ class UcbTunedLearner(Learner[Context, Action]):
             action: The action that was selected in the context. See the base class for more information.
             reward: The reward that was gained from the action. See the base class for more information.
         """
-        if action not in self._s:
-            self._s[action] = 1
-        else:
-            self._s[action] += 1
 
         if action not in self._m:
             self._m[action] = reward
@@ -515,10 +441,12 @@ class VowpalLearner(Learner[Context, Action]):
             self._exploration = exploration
 
         self._probs: Dict[Key, float] = {}
-        self._actions = self._new_actions(learning)
+        self._actions = self._new_actions(self._learning)
+
+        self._flags = kwargs.get('flags', '')
+
+        self._vw = VW.pyvw_Wrapper(self._learning.formatter, seed=kwargs.get('seed', None))
 
-        self._vw = VW.pyvw_Wrapper(learning.formatter, seed=kwargs.get('seed', None))
-    
     @property
     def family(self) -> str:
         """The family of the learner.
@@ -533,7 +461,7 @@ class VowpalLearner(Learner[Context, Action]):
         
         See the base class for more information
         """        
-        return {**self._learning.params(), **self._exploration.params()}
+        return {**self._learning.params(), **self._exploration.params()}        
 
     def choose(self, key: Key, context: Context, actions: Sequence[Action]) -> Choice:
         """Choose an action according to the VowpalWabbit parameters passed into the contructor.
@@ -549,7 +477,7 @@ class VowpalLearner(Learner[Context, Action]):
         """
 
         if not self._vw.created:
-            self._vw.create(self._learning.flags(actions) + " " + self._exploration.flags())        
+            self._vw.create(self._learning.flags(actions) + " " + self._exploration.flags() + " " + self._flags)
 
         choice, prob = self._vw.choose(context, actions)
 
@@ -593,4 +521,110 @@ class VowpalLearner(Learner[Context, Action]):
         if isinstance(self._actions, collections.MutableMapping) :
             return self._actions.pop(key)
         else:
-            return self._actions
\ No newline at end of file
+            return self._actions
+
+class CorralLearner(Learner[Context, Action]):
+
+    def __init__(self, base_learners: Sequence[Learner[Context,Action]], eta: float, T: float = math.inf, seed: int = None) -> None:
+        
+        self._base_learners = base_learners
+
+        M = len(self._base_learners)
+
+        self._M     = M
+        self._gamma = 1/T
+        self._beta  = 1/math.exp(1/math.log(T))
+
+        self._eta_init = eta
+        self._etas     = [ eta ] * M
+        self._rhos     = [ float(2*M) ] * M
+        self._ps       = [ 1/M ] * M
+        self._p_bars   = [ 1/M ] * M
+
+        self._random   = CobaRandom(seed)
+        self._chosen_i: Dict[Key,int] = {}
+
+    @property
+    def family(self) -> str:
+        """The family of the learner.
+
+        See the base class for more information
+        """
+        return "corral"
+    
+    @property
+    def params(self) -> Dict[str, Any]:
+        """The parameters of the learner.
+        
+        See the base class for more information
+        """        
+        return {"eta": self._eta_init, "B": [ b.family for b in self._base_learners ] }
+
+    def choose(self, key: Key, context: _C_in, actions: Sequence[_A_in]) -> Choice:
+
+        thetas = [ base_algorithm.choose(key, context, actions) for base_algorithm in self._base_learners ]
+
+        i = self._random.choice(range(self._M), self._p_bars)
+
+        self._chosen_i[key] = i
+
+        return thetas[i]
+
+    def learn(self, key: Key, context: _C_in, action: _A_in, reward: Reward) -> None:
+
+        loss = 1-reward # this assumes reward \in [0,1]
+        chosen_i = self._chosen_i.pop(key)
+
+        rewards = [ reward/self._p_bars[chosen_i] * int(i == chosen_i) for i in range(self._M)]
+        losses  = [ loss/self._p_bars[chosen_i] * int(i == chosen_i) for i in range(self._M)]
+
+        for learner,reward in zip(self._base_learners, rewards):
+            learner.learn(key, context, action, reward)
+
+        self._ps     = list(self._log_barrier_omd(losses))
+        self._p_bars = [ (1-self._gamma)*p + self._gamma*1/self._M for p in self._ps ]
+
+        for i in range(self._M):
+            if 1/self._p_bars[i] > self._rhos[i]:
+                self._rhos[i] = 2/self._p_bars[i]
+                self._etas[i] *= self._beta
+
+    def _log_barrier_omd(self, losses) -> Sequence[float]:
+
+        f  = lambda l: float(sum( [ 1/((1/p) + eta*(loss-l)) for p, eta, loss in zip(self._ps, self._etas, losses)]))
+        df = lambda l: float(sum( [ eta/((1/p) + eta*(loss-l))**2 for p, eta, loss in zip(self._ps, self._etas, losses)]))
+
+        min_loss = min(losses)
+        max_loss = max(losses)
+
+        if max_loss - 1 < .0001:
+            l = float(1)
+        elif max_loss > 2000:
+            l = float(0)
+        else:
+            l = (min_loss + max_loss)/2
+
+        i = 0
+
+        while True:
+
+            i += 1
+
+            #try:
+            if l < min_loss or l > max_loss:
+                l = min_loss + self._random.random() * (max_loss-min_loss)
+
+            if df(l) == 0:
+                print('what happened?')
+
+            l = l - (f(l)-1)/df(l)
+
+            if round(f(l)-1,5) == 0:
+                break
+            #except:
+            #    l = min_loss + self._random.random() * (max_loss-min_loss)
+
+            if (i % 30000) == 0:
+                print(i)
+
+        return [ 1/((1/p) + eta*(loss-l)) for p, eta, loss in zip(self._ps, self._etas, losses)]
\ No newline at end of file
